@@ -3,12 +3,12 @@
  * copy.c
  *		Implements the COPY utility command
  *
- * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/commands/copy.c,v 1.295 2008/01/01 19:45:48 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/commands/copy.c,v 1.312 2009/06/11 14:48:55 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -40,6 +40,7 @@
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
 #include "utils/memutils.h"
+#include "utils/snapmgr.h"
 
 
 #define ISOCTAL(c) (((c) >= '0') && ((c) <= '7'))
@@ -176,8 +177,11 @@ typedef struct
  * They are macros because they often do continue/break control and to avoid
  * function call overhead in tight COPY loops.
  *
- * We must use "if (1)" because "do {} while(0)" overrides the continue/break
- * processing.	See http://www.cit.gu.edu.au/~anthony/info/C/C.macros.
+ * We must use "if (1)" because the usual "do {...} while(0)" wrapper would
+ * prevent the continue/break processing from working.	We end the "if (1)"
+ * with "else ((void) 0)" to ensure the "if" does not unintentionally match
+ * any "else" in the calling code, and to avoid any compiler warnings about
+ * empty statements.  See http://www.cit.gu.edu.au/~anthony/info/C/C.macros.
  */
 
 /*
@@ -193,8 +197,7 @@ if (1) \
 		need_data = true; \
 		continue; \
 	} \
-} else
-
+} else ((void) 0)
 
 /* This consumes the remainder of the buffer and breaks */
 #define IF_NEED_REFILL_AND_EOF_BREAK(extralen) \
@@ -208,8 +211,7 @@ if (1) \
 		result = true; \
 		break; \
 	} \
-} else
-
+} else ((void) 0)
 
 /*
  * Transfer any approved data to line_buf; must do this to be sure
@@ -225,7 +227,7 @@ if (1) \
 							   raw_buf_ptr - cstate->raw_buf_index); \
 		cstate->raw_buf_index = raw_buf_ptr; \
 	} \
-} else
+} else ((void) 0)
 
 /* Undo any read-ahead and jump out of the block. */
 #define NO_END_OF_COPY_GOTO \
@@ -233,8 +235,7 @@ if (1) \
 { \
 	raw_buf_ptr = prev_raw_ptr + 1; \
 	goto not_end_of_copy; \
-} else
-
+} else ((void) 0)
 
 static const char BinarySignature[11] = "PGCOPY\n\377\r\n\0";
 
@@ -710,7 +711,7 @@ CopyLoadRawBuf(CopyState cstate)
  * or write to a file.
  *
  * Do not allow the copy if user doesn't have proper permission to access
- * the table.
+ * the table or the specifically requested columns.
  */
 uint64
 DoCopy(const CopyStmt *stmt, const char *queryString)
@@ -722,7 +723,8 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 	List	   *force_quote = NIL;
 	List	   *force_notnull = NIL;
 	AclMode		required_access = (is_from ? ACL_INSERT : ACL_SELECT);
-	AclResult	aclresult;
+	AclMode		relPerms;
+	AclMode		remainingPerms;
 	ListCell   *option;
 	TupleDesc	tupDesc;
 	int			num_phys_attrs;
@@ -853,11 +855,11 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 			cstate->escape = cstate->quote;
 	}
 
-	/* Only single-character delimiter strings are supported. */
+	/* Only single-byte delimiter strings are supported. */
 	if (strlen(cstate->delim) != 1)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("COPY delimiter must be a single ASCII character")));
+			  errmsg("COPY delimiter must be a single one-byte character")));
 
 	/* Disallow end-of-line characters */
 	if (strchr(cstate->delim, '\r') != NULL ||
@@ -877,8 +879,8 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 	 * backslash because it would be ambiguous.  We can't allow the other
 	 * cases because data characters matching the delimiter must be
 	 * backslashed, and certain backslash combinations are interpreted
-	 * non-literally by COPY IN.  Disallowing all lower case ASCII letters
-	 * is more than strictly necessary, but seems best for consistency and
+	 * non-literally by COPY IN.  Disallowing all lower case ASCII letters is
+	 * more than strictly necessary, but seems best for consistency and
 	 * future-proofing.  Likewise we disallow all digits though only octal
 	 * digits are actually dangerous.
 	 */
@@ -904,7 +906,7 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 	if (cstate->csv_mode && strlen(cstate->quote) != 1)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("COPY quote must be a single ASCII character")));
+				 errmsg("COPY quote must be a single one-byte character")));
 
 	if (cstate->csv_mode && cstate->delim[0] == cstate->quote[0])
 		ereport(ERROR,
@@ -920,7 +922,7 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 	if (cstate->csv_mode && strlen(cstate->escape) != 1)
 		ereport(ERROR,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("COPY escape must be a single ASCII character")));
+				 errmsg("COPY escape must be a single one-byte character")));
 
 	/* Check force_quote */
 	if (!cstate->csv_mode && force_quote != NIL)
@@ -972,17 +974,34 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 		cstate->rel = heap_openrv(stmt->relation,
 							 (is_from ? RowExclusiveLock : AccessShareLock));
 
+		tupDesc = RelationGetDescr(cstate->rel);
+
 		/* Check relation permissions. */
-		aclresult = pg_class_aclcheck(RelationGetRelid(cstate->rel),
-									  GetUserId(),
-									  required_access);
-		if (aclresult != ACLCHECK_OK)
-			aclcheck_error(aclresult, ACL_KIND_CLASS,
-						   RelationGetRelationName(cstate->rel));
+		relPerms = pg_class_aclmask(RelationGetRelid(cstate->rel), GetUserId(),
+									required_access, ACLMASK_ALL);
+		remainingPerms = required_access & ~relPerms;
+		if (remainingPerms != 0)
+		{
+			/* We don't have table permissions, check per-column permissions */
+			List	   *attnums;
+			ListCell   *cur;
+
+			attnums = CopyGetAttnums(tupDesc, cstate->rel, attnamelist);
+			foreach(cur, attnums)
+			{
+				int			attnum = lfirst_int(cur);
+
+				if (pg_attribute_aclcheck(RelationGetRelid(cstate->rel),
+										  attnum,
+										  GetUserId(),
+										  remainingPerms) != ACLCHECK_OK)
+					aclcheck_error(ACLCHECK_NO_PRIV, ACL_KIND_CLASS,
+								   RelationGetRelationName(cstate->rel));
+			}
+		}
 
 		/* check read-only transaction */
-		if (XactReadOnly && is_from &&
-			!isTempNamespace(RelationGetNamespace(cstate->rel)))
+		if (XactReadOnly && is_from && !cstate->rel->rd_islocaltemp)
 			ereport(ERROR,
 					(errcode(ERRCODE_READ_ONLY_SQL_TRANSACTION),
 					 errmsg("transaction is read-only")));
@@ -993,8 +1012,6 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 					(errcode(ERRCODE_UNDEFINED_COLUMN),
 					 errmsg("table \"%s\" does not have OIDs",
 							RelationGetRelationName(cstate->rel))));
-
-		tupDesc = RelationGetDescr(cstate->rel);
 	}
 	else
 	{
@@ -1043,21 +1060,19 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 		plan = planner(query, 0, NULL);
 
 		/*
-		 * Update snapshot command ID to ensure this query sees results of any
-		 * previously executed queries.  (It's a bit cheesy to modify
-		 * ActiveSnapshot without making a copy, but for the limited ways in
-		 * which COPY can be invoked, I think it's OK, because the active
-		 * snapshot shouldn't be shared with anything else anyway.)
+		 * Use a snapshot with an updated command ID to ensure this query sees
+		 * results of any previously executed queries.
 		 */
-		ActiveSnapshot->curcid = GetCurrentCommandId(false);
+		PushUpdatedSnapshot(GetActiveSnapshot());
 
 		/* Create dest receiver for COPY OUT */
-		dest = CreateDestReceiver(DestCopyOut, NULL);
+		dest = CreateDestReceiver(DestCopyOut);
 		((DR_copy *) dest)->cstate = cstate;
 
 		/* Create a QueryDesc requesting no output */
-		cstate->queryDesc = CreateQueryDesc(plan,
-											ActiveSnapshot, InvalidSnapshot,
+		cstate->queryDesc = CreateQueryDesc(plan, queryString,
+											GetActiveSnapshot(),
+											InvalidSnapshot,
 											dest, NULL, false);
 
 		/*
@@ -1160,6 +1175,7 @@ DoCopy(const CopyStmt *stmt, const char *queryString)
 		/* Close down the query and free resources. */
 		ExecutorEnd(cstate->queryDesc);
 		FreeQueryDesc(cstate->queryDesc);
+		PopActiveSnapshot();
 	}
 
 	/* Clean up storage (probably not really necessary) */
@@ -1389,7 +1405,7 @@ CopyTo(CopyState cstate)
 		values = (Datum *) palloc(num_phys_attrs * sizeof(Datum));
 		nulls = (bool *) palloc(num_phys_attrs * sizeof(bool));
 
-		scandesc = heap_beginscan(cstate->rel, ActiveSnapshot, 0, NULL);
+		scandesc = heap_beginscan(cstate->rel, GetActiveSnapshot(), 0, NULL);
 
 		while ((tuple = heap_getnext(scandesc, ForwardScanDirection)) != NULL)
 		{
@@ -1639,7 +1655,7 @@ CopyFrom(CopyState cstate)
 	int			i;
 	Oid			in_func_oid;
 	Datum	   *values;
-	char	   *nulls;
+	bool	   *nulls;
 	int			nfields;
 	char	  **field_strings;
 	bool		done = false;
@@ -1654,8 +1670,8 @@ CopyFrom(CopyState cstate)
 	MemoryContext oldcontext = CurrentMemoryContext;
 	ErrorContextCallback errcontext;
 	CommandId	mycid = GetCurrentCommandId(true);
-	bool		use_wal = true; /* by default, use WAL logging */
-	bool		use_fsm = true; /* by default, use FSM for free space */
+	int			hi_options = 0; /* start with default heap_insert options */
+	BulkInsertState bistate;
 
 	Assert(cstate->rel);
 
@@ -1708,9 +1724,9 @@ CopyFrom(CopyState cstate)
 	if (cstate->rel->rd_createSubid != InvalidSubTransactionId ||
 		cstate->rel->rd_newRelfilenodeSubid != InvalidSubTransactionId)
 	{
-		use_fsm = false;
+		hi_options |= HEAP_INSERT_SKIP_FSM;
 		if (!XLogArchivingActive())
-			use_wal = false;
+			hi_options |= HEAP_INSERT_SKIP_WAL;
 	}
 
 	if (pipe)
@@ -1873,7 +1889,7 @@ CopyFrom(CopyState cstate)
 	}
 
 	values = (Datum *) palloc(num_phys_attrs * sizeof(Datum));
-	nulls = (char *) palloc(num_phys_attrs * sizeof(char));
+	nulls = (bool *) palloc(num_phys_attrs * sizeof(bool));
 
 	/* create workspace for CopyReadAttributes results */
 	nfields = file_has_oids ? (attr_count + 1) : attr_count;
@@ -1887,6 +1903,8 @@ CopyFrom(CopyState cstate)
 	cstate->cur_attname = NULL;
 	cstate->cur_attval = NULL;
 
+	bistate = GetBulkInsertState();
+
 	/* Set up callback to identify error line number */
 	errcontext.callback = copy_in_error_callback;
 	errcontext.arg = (void *) cstate;
@@ -1917,7 +1935,7 @@ CopyFrom(CopyState cstate)
 
 		/* Initialize all values for row to NULL */
 		MemSet(values, 0, num_phys_attrs * sizeof(Datum));
-		MemSet(nulls, 'n', num_phys_attrs * sizeof(char));
+		MemSet(nulls, true, num_phys_attrs * sizeof(bool));
 
 		if (!cstate->binary)
 		{
@@ -1999,7 +2017,7 @@ CopyFrom(CopyState cstate)
 											  typioparams[m],
 											  attr[m]->atttypmod);
 				if (string != NULL)
-					nulls[m] = ' ';
+					nulls[m] = false;
 				cstate->cur_attname = NULL;
 				cstate->cur_attval = NULL;
 			}
@@ -2055,8 +2073,7 @@ CopyFrom(CopyState cstate)
 													&in_functions[m],
 													typioparams[m],
 													attr[m]->atttypmod,
-													&isnull);
-				nulls[m] = isnull ? 'n' : ' ';
+													&nulls[m]);
 				cstate->cur_attname = NULL;
 			}
 		}
@@ -2069,13 +2086,11 @@ CopyFrom(CopyState cstate)
 		for (i = 0; i < num_defaults; i++)
 		{
 			values[defmap[i]] = ExecEvalExpr(defexprs[i], econtext,
-											 &isnull, NULL);
-			if (!isnull)
-				nulls[defmap[i]] = ' ';
+											 &nulls[defmap[i]], NULL);
 		}
 
 		/* And now we can form the input tuple. */
-		tuple = heap_formtuple(tupDesc, values, nulls);
+		tuple = heap_form_tuple(tupDesc, values, nulls);
 
 		if (cstate->oids && file_has_oids)
 			HeapTupleSetOid(tuple, loaded_oid);
@@ -2112,7 +2127,7 @@ CopyFrom(CopyState cstate)
 				ExecConstraints(resultRelInfo, slot, estate);
 
 			/* OK, store the tuple and create index entries for it */
-			heap_insert(cstate->rel, tuple, mycid, use_wal, use_fsm);
+			heap_insert(cstate->rel, tuple, mycid, hi_options, bistate);
 
 			if (resultRelInfo->ri_NumIndices > 0)
 				ExecInsertIndexTuples(slot, &(tuple->t_self), estate, false);
@@ -2132,6 +2147,8 @@ CopyFrom(CopyState cstate)
 	/* Done, clean up */
 	error_context_stack = errcontext.previous;
 
+	FreeBulkInsertState(bistate);
+
 	MemoryContextSwitchTo(oldcontext);
 
 	/* Execute AFTER STATEMENT insertion triggers */
@@ -2168,7 +2185,7 @@ CopyFrom(CopyState cstate)
 	 * If we skipped writing WAL, then we need to sync the heap (but not
 	 * indexes since those use WAL anyway)
 	 */
-	if (!use_wal)
+	if (hi_options & HEAP_INSERT_SKIP_WAL)
 		heap_sync(cstate->rel);
 }
 
@@ -2437,10 +2454,10 @@ CopyReadLineText(CopyState cstate)
 						ereport(ERROR,
 								(errcode(ERRCODE_BAD_COPY_FILE_FORMAT),
 								 !cstate->csv_mode ?
-								 errmsg("literal carriage return found in data") :
-								 errmsg("unquoted carriage return found in data"),
+							errmsg("literal carriage return found in data") :
+							errmsg("unquoted carriage return found in data"),
 								 !cstate->csv_mode ?
-								 errhint("Use \"\\r\" to represent carriage return.") :
+						errhint("Use \"\\r\" to represent carriage return.") :
 								 errhint("Use quoted CSV field to represent carriage return.")));
 
 					/*
@@ -2457,7 +2474,7 @@ CopyReadLineText(CopyState cstate)
 						 errmsg("literal carriage return found in data") :
 						 errmsg("unquoted carriage return found in data"),
 						 !cstate->csv_mode ?
-						 errhint("Use \"\\r\" to represent carriage return.") :
+					   errhint("Use \"\\r\" to represent carriage return.") :
 						 errhint("Use quoted CSV field to represent carriage return.")));
 			/* If reach here, we have found the line terminator */
 			break;
@@ -2474,7 +2491,7 @@ CopyReadLineText(CopyState cstate)
 						 errmsg("unquoted newline found in data"),
 						 !cstate->csv_mode ?
 						 errhint("Use \"\\n\" to represent newline.") :
-						 errhint("Use quoted CSV field to represent newline.")));
+					 errhint("Use quoted CSV field to represent newline.")));
 			cstate->eol_type = EOL_NL;	/* in case not set yet */
 			/* If reach here, we have found the line terminator */
 			break;
@@ -2701,7 +2718,7 @@ CopyReadAttributesText(CopyState cstate, int maxfields, char **fieldvals)
 		char	   *start_ptr;
 		char	   *end_ptr;
 		int			input_len;
-		bool		saw_high_bit = false;
+		bool		saw_non_ascii = false;
 
 		/* Make sure space remains in fieldvals[] */
 		if (fieldno >= maxfields)
@@ -2713,7 +2730,17 @@ CopyReadAttributesText(CopyState cstate, int maxfields, char **fieldvals)
 		start_ptr = cur_ptr;
 		fieldvals[fieldno] = output_ptr;
 
-		/* Scan data for field */
+		/*
+		 * Scan data for field.
+		 *
+		 * Note that in this loop, we are scanning to locate the end of field
+		 * and also speculatively performing de-escaping.  Once we find the
+		 * end-of-field, we can match the raw field contents against the null
+		 * marker string.  Only after that comparison fails do we know that
+		 * de-escaping is actually the right thing to do; therefore we *must
+		 * not* throw any syntax errors before we've done the null-marker
+		 * check.
+		 */
 		for (;;)
 		{
 			char		c;
@@ -2766,8 +2793,8 @@ CopyReadAttributesText(CopyState cstate, int maxfields, char **fieldvals)
 								}
 							}
 							c = val & 0377;
-							if (IS_HIGHBIT_SET(c))
-								saw_high_bit = true;
+							if (c == '\0' || IS_HIGHBIT_SET(c))
+								saw_non_ascii = true;
 						}
 						break;
 					case 'x':
@@ -2791,8 +2818,8 @@ CopyReadAttributesText(CopyState cstate, int maxfields, char **fieldvals)
 									}
 								}
 								c = val & 0xff;
-								if (IS_HIGHBIT_SET(c))
-									saw_high_bit = true;
+								if (c == '\0' || IS_HIGHBIT_SET(c))
+									saw_non_ascii = true;
 							}
 						}
 						break;
@@ -2826,26 +2853,29 @@ CopyReadAttributesText(CopyState cstate, int maxfields, char **fieldvals)
 			*output_ptr++ = c;
 		}
 
-		/* Terminate attribute value in output area */
-		*output_ptr++ = '\0';
-
-		/*
-		 * If we de-escaped a char with the high bit set, make sure we still
-		 * have valid data for the db encoding. Avoid calling strlen here for
-		 * the sake of efficiency.
-		 */
-		if (saw_high_bit)
-		{
-			char	   *fld = fieldvals[fieldno];
-
-			pg_verifymbstr(fld, output_ptr - (fld + 1), false);
-		}
-
 		/* Check whether raw input matched null marker */
 		input_len = end_ptr - start_ptr;
 		if (input_len == cstate->null_print_len &&
 			strncmp(start_ptr, cstate->null_print, input_len) == 0)
 			fieldvals[fieldno] = NULL;
+		else
+		{
+			/*
+			 * At this point we know the field is supposed to contain data.
+			 *
+			 * If we de-escaped any non-7-bit-ASCII chars, make sure the
+			 * resulting string is valid data for the db encoding.
+			 */
+			if (saw_non_ascii)
+			{
+				char	   *fld = fieldvals[fieldno];
+
+				pg_verifymbstr(fld, output_ptr - fld, false);
+			}
+		}
+
+		/* Terminate attribute value in output area */
+		*output_ptr++ = '\0';
 
 		fieldno++;
 		/* Done if we hit EOL instead of a delim */
@@ -2913,7 +2943,6 @@ CopyReadAttributesCSV(CopyState cstate, int maxfields, char **fieldvals)
 	for (;;)
 	{
 		bool		found_delim = false;
-		bool		in_quote = false;
 		bool		saw_quote = false;
 		char	   *start_ptr;
 		char	   *end_ptr;
@@ -2929,72 +2958,88 @@ CopyReadAttributesCSV(CopyState cstate, int maxfields, char **fieldvals)
 		start_ptr = cur_ptr;
 		fieldvals[fieldno] = output_ptr;
 
-		/* Scan data for field */
+		/*
+		 * Scan data for field,
+		 *
+		 * The loop starts in "not quote" mode and then toggles between that
+		 * and "in quote" mode. The loop exits normally if it is in "not
+		 * quote" mode and a delimiter or line end is seen.
+		 */
 		for (;;)
 		{
 			char		c;
 
-			end_ptr = cur_ptr;
-			if (cur_ptr >= line_end_ptr)
-				break;
-			c = *cur_ptr++;
-			/* unquoted field delimiter */
-			if (c == delimc && !in_quote)
+			/* Not in quote */
+			for (;;)
 			{
-				found_delim = true;
-				break;
-			}
-			/* start of quoted field (or part of field) */
-			if (c == quotec && !in_quote)
-			{
-				saw_quote = true;
-				in_quote = true;
-				continue;
-			}
-			/* escape within a quoted field */
-			if (c == escapec && in_quote)
-			{
-				/*
-				 * peek at the next char if available, and escape it if it is
-				 * an escape char or a quote char
-				 */
-				if (cur_ptr < line_end_ptr)
+				end_ptr = cur_ptr;
+				if (cur_ptr >= line_end_ptr)
+					goto endfield;
+				c = *cur_ptr++;
+				/* unquoted field delimiter */
+				if (c == delimc)
 				{
-					char		nextc = *cur_ptr;
+					found_delim = true;
+					goto endfield;
+				}
+				/* start of quoted field (or part of field) */
+				if (c == quotec)
+				{
+					saw_quote = true;
+					break;
+				}
+				/* Add c to output string */
+				*output_ptr++ = c;
+			}
 
-					if (nextc == escapec || nextc == quotec)
+			/* In quote */
+			for (;;)
+			{
+				end_ptr = cur_ptr;
+				if (cur_ptr >= line_end_ptr)
+					ereport(ERROR,
+							(errcode(ERRCODE_BAD_COPY_FILE_FORMAT),
+							 errmsg("unterminated CSV quoted field")));
+
+				c = *cur_ptr++;
+
+				/* escape within a quoted field */
+				if (c == escapec)
+				{
+					/*
+					 * peek at the next char if available, and escape it if it
+					 * is an escape char or a quote char
+					 */
+					if (cur_ptr < line_end_ptr)
 					{
-						*output_ptr++ = nextc;
-						cur_ptr++;
-						continue;
+						char		nextc = *cur_ptr;
+
+						if (nextc == escapec || nextc == quotec)
+						{
+							*output_ptr++ = nextc;
+							cur_ptr++;
+							continue;
+						}
 					}
 				}
-			}
 
-			/*
-			 * end of quoted field. Must do this test after testing for escape
-			 * in case quote char and escape char are the same (which is the
-			 * common case).
-			 */
-			if (c == quotec && in_quote)
-			{
-				in_quote = false;
-				continue;
-			}
+				/*
+				 * end of quoted field. Must do this test after testing for
+				 * escape in case quote char and escape char are the same
+				 * (which is the common case).
+				 */
+				if (c == quotec)
+					break;
 
-			/* Add c to output string */
-			*output_ptr++ = c;
+				/* Add c to output string */
+				*output_ptr++ = c;
+			}
 		}
+endfield:
 
 		/* Terminate attribute value in output area */
 		*output_ptr++ = '\0';
 
-		/* Shouldn't still be in quote mode */
-		if (in_quote)
-			ereport(ERROR,
-					(errcode(ERRCODE_BAD_COPY_FILE_FORMAT),
-					 errmsg("unterminated CSV quoted field")));
-
 		/* Check whether raw input matched null marker */
 		input_len = end_ptr - start_ptr;
 		if (!saw_quote && input_len == cstate->null_print_len &&
@@ -3113,11 +3158,11 @@ CopyAttributeOutText(CopyState cstate, char *string)
 			if ((unsigned char) c < (unsigned char) 0x20)
 			{
 				/*
-				 * \r and \n must be escaped, the others are traditional.
-				 * We prefer to dump these using the C-like notation, rather
-				 * than a backslash and the literal character, because it
-				 * makes the dump file a bit more proof against Microsoftish
-				 * data mangling.
+				 * \r and \n must be escaped, the others are traditional. We
+				 * prefer to dump these using the C-like notation, rather than
+				 * a backslash and the literal character, because it makes the
+				 * dump file a bit more proof against Microsoftish data
+				 * mangling.
 				 */
 				switch (c)
 				{
@@ -3151,7 +3196,7 @@ CopyAttributeOutText(CopyState cstate, char *string)
 				DUMPSOFAR();
 				CopySendChar(cstate, '\\');
 				CopySendChar(cstate, c);
-				start = ++ptr;			/* do not include char in next run */
+				start = ++ptr;	/* do not include char in next run */
 			}
 			else if (c == '\\' || c == delimc)
 			{
@@ -3173,11 +3218,11 @@ CopyAttributeOutText(CopyState cstate, char *string)
 			if ((unsigned char) c < (unsigned char) 0x20)
 			{
 				/*
-				 * \r and \n must be escaped, the others are traditional.
-				 * We prefer to dump these using the C-like notation, rather
-				 * than a backslash and the literal character, because it
-				 * makes the dump file a bit more proof against Microsoftish
-				 * data mangling.
+				 * \r and \n must be escaped, the others are traditional. We
+				 * prefer to dump these using the C-like notation, rather than
+				 * a backslash and the literal character, because it makes the
+				 * dump file a bit more proof against Microsoftish data
+				 * mangling.
 				 */
 				switch (c)
 				{
@@ -3211,7 +3256,7 @@ CopyAttributeOutText(CopyState cstate, char *string)
 				DUMPSOFAR();
 				CopySendChar(cstate, '\\');
 				CopySendChar(cstate, c);
-				start = ++ptr;			/* do not include char in next run */
+				start = ++ptr;	/* do not include char in next run */
 			}
 			else if (c == '\\' || c == delimc)
 			{
