@@ -1,5 +1,5 @@
 /*
- * $PostgreSQL: pgsql/src/include/regex/regerrs.h,v 1.3.6.1 2008/01/03 20:48:57 tgl Exp $
+ * $PostgreSQL: pgsql/src/include/regex/regerrs.h,v 1.5 2008/01/03 20:47:55 tgl Exp $
  */
 
 {
@@ -59,7 +59,7 @@
 },
 
 {
-	REG_ASSERT, "REG_ASSERT", "\"can't happen\" -- you found a bug"
+	REG_ASSERT, "REG_ASSERT", "\"cannot happen\" -- you found a bug"
 },
 
 {
@@ -77,3 +77,7 @@
 {
 	REG_ETOOBIG, "REG_ETOOBIG", "nfa has too many states"
 },
+
+{
+	REG_ECOLORS, "REG_ECOLORS", "too many colors"
+},
